/* Copyright (c) 2006 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

/* $Header: /code/qtpod/qtpod.cpp,v 1.25 2006/07/31 15:22:17 doj Exp $ */

#include <iostream>

#include <assert.h>

#include <qapplication.h>
#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qgroupbox.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qscrollview.h>
#include <qsettings.h>
#include <qslider.h>
#include <qsocketnotifier.h>
#include <qtimer.h>
#include <qvbox.h>
#include <qwidgetlist.h>

#include "qtpod.hpp"
#include "midi.hpp"

using namespace std;

const int TIMER_DELAY=250;

#define sizeof_array(x) (sizeof(x)/sizeof(x[0]))

#include <endian.h>
#if __BYTE_ORDER == __LITTLE_ENDIAN
#define le16(x) x
#else
inline uint16_t le16(const uint16_t x) { return ((x&0xFF)<<8) + ((x&0xFF00)>>8); }
#endif

static const char* presetname_prefix[] = {
  "[ManualMode]",
  "1A ", "1B ", "1C ", "1D ",
  "2A ", "2B ", "2C ", "2D ",
  "3A ", "3B ", "3C ", "3D ",
  "4A ", "4B ", "4C ", "4D ",
  "5A ", "5B ", "5C ", "5D ",
  "6A ", "6B ", "6C ", "6D ",
  "7A ", "7B ", "7C ", "7D ",
  "8A ", "8B ", "8C ", "8D ",
  "9A ", "9B ", "9C ", "9D ",
  "[Tuner]",
};

static const char* ampname[] = {
  "Tube Preamp",
  "Line 6 Clean",
  "Line 6 Crunch",
  "Line 6 Drive",
  "Line 6 Layer",
  "Small Tweed",
  "Tweed Blues",
  "Black Panel",
  "Modern Class A",
  "Brit Class A",
  "Brit Blues",
  "Brit Classic",
  "Brit Hi Gain",
  "Rectified",
  "Modern Hi Gain",
  "Fuzz Box",
  "Jazz Clean",
  "Boutique 1",
  "Boutique 2",
  "Brit Class A 2",
  "Brit Class A 3",
  "Small Tweed 2",
  "Black Panel 2",
  "Boutique 3",
  "California Crunch 1",
  "California Crunch 2",
  "Rectified 2",
  "Modern Hi Gain 2",
  "Line 6 Twang",
  "Line 6 Crunch 2",
  "Line 6 Blues",
  "Line 6 Insane",
};

static const char* cabname[] = {
  "1x8  '60 Fender Tweed Champ",
  "1x12 '52 Fender Tweed Deluxe",
  "1x12 '60 Vox AC15",
  "1x12 '64 Fender Blackface Deluxe",
  "1x12 '98 Line6 Flextone",
  "2x12 '65 Fender Blackface Twin",
  "2x12 '67 Vox AC30",
  "2x12 '65 Matchless Chieftain",
  "2x12 '98 Line 6 Custom 2x12",
  "4x10 '59 Fender Bassman",
  "4x10 '98 Line 6 Custom 4x10",
  "4x12 '96 Marshall with V30s",
  "4x12 '78 Marshall with stock 70",
  "4x12 '97 Marshall with Greenbacks",
  "4x12 '98 Line 6 Custom 4x12",
  "No Cabinet Emulation",
};

static const char* fxname[] = {
  "Choruss 2",
  "Flanger 1",
  "Rotary Speaker",
  "Flanger 2",
  "Delay+Chorus1",
  "Delay+Tremolo",
  "Delay",
  "Delay+Compressor",
  "Chorus 1",
  "Tremolo",
  "Bypass",
  "Compressor",
  "Delay+Chorus2",
  "Delay+Flanger1",
  "Delay+Swell",
  "Delay+Flanger2",
};

void QTPOD::parsecontroller(const int c, const int v)
{
  switch(c)
    {
      // Switches
    case 25: sw_distortion->setChecked(v); break;
    case 26: sw_drive->setChecked(v); break;
    case 27: sw_eq->setChecked(v); break;
    case 28: sw_delay->setChecked(v); break;
    case 50: sw_tremolo->setChecked(v); break;
    case 36: sw_reverb->setChecked(v); break;
    case 22: sw_noise->setChecked(v); break;
    case 73: sw_bright->setChecked(v); break;

      // Preamp
    case 12: preamp_ampmodel->setCurrentItem(v); updatebrightsw(v); break;
    case 13: preamp_drive->setValue(v); break;
    case 20: preamp_drive2->setValue(v); break;
    case 14: preamp_bass->setValue(v); break;
    case 15: preamp_mid->setValue(v); break;
    case 16: preamp_treble->setValue(v); break;
    case 21: preamp_presence->setValue(v); break;
    case 17: preamp_chanvol->setValue(v); break;

      // NoiseGate
    case 23: ng_threshold->setValue(v); break;
    case 24: ng_decay->setValue(v); break;

      // WahWah
    case 4:  wah_level->setValue(v); break;
    case 43: wah_enable->setChecked(v); break;
    case 44: wah_botfreq->setValue(v); break;
    case 45: wah_topfreq->setValue(v); break;

      // Vol Pedal
    case  7: volpedal_level->setValue(v); break;
    case 46: volpedal_min->setValue(v); break;
    case 47: volpedal_pos->setChecked(v); break;

      // Delay
    case 30: delay_time1->setValue( (delay_time1->value() & 0x7F) | ((v&0x7F)<<7) ); break;
    case 62: delay_time1->setValue( (delay_time1->value() & 0xFF80) | (v&0x7F) ); break;
    case 32: delay_feedback1->setValue(v); break;
    case 34: delay_level1->setValue(v); break;

      // Reverb
    case 37: reverb_type->setButton(v/64); break;
    case 38: reverb_decay->setValue(v); break;
    case 39: reverb_tone->setValue(v); break;
    case 40: reverb_diffusion->setValue(v); break;
    case 41: reverb_density->setValue(v); break;
    case 18: reverb_level->setValue(v); break;

      // Cabinet
    case 71: cab_type->setCurrentItem(v); break;
    case 72: cab_air->setValue(v); break;

      // FX Config
    case 19: fx_type->setCurrentItem(v); break;
    case  1: fx_tweak->setValue(v); break;

      // FX
    case 49: swell_time->setValue(v); break;
    case 42: comp_ration->setValue(v); break;
    case 51: chorus_speed->setValue(v); break;
    case 52: chorus_depth->setValue(v); break;
    case 53: chorus_feedback->setValue(v); break;
    case 54: chorus_delay->setValue(v); break;
    case 55: rotary_curr->setChecked(v); break;
    case 56: rotary_fast->setValue(v); break;
    case 57: rotary_slow->setValue(v); break;
    case 58: tremolo_speed->setValue(v); break;
    case 59: tremolo_depth->setValue(v); break;

      // Dig Gain
    case 9:  dig_gain->setValue(v); break;

      // Tap Tempo
    case 64: taptempo(v); break;

    default: cerr << "controller " << c << " " << v << endl; break;
    }
}

void QTPOD::parsesysex(const uint8_t *buf, const unsigned len)
{
#if 0 //def DOJDEBUG
  fprintf(stderr, "parsesysex:%i: ", len);
  for(unsigned i=0; i!=len; ++i)
    fprintf(stderr, "%02X ", buf[i]);
  fprintf(stderr, "\n");
#endif

  podfound=true;

  // Universal Device Inquiery
  if(len>=17 && buf[0]==0xF0 && buf[1]==0x7E && buf[3]==6 && buf[4]==2 && buf[16]==0xF7)
    {
#pragma pack(push, 1)
      struct udi {
	uint8_t id[5];
	uint8_t line6id[3];
	uint16_t family;
	uint16_t product;
	char version[4];
	uint8_t eox;
      };
#pragma pack(pop)
      const udi *b=reinterpret_cast<const udi*>(buf);
      podname.truncate(0);
      if(b->line6id[0]==0 && b->line6id[1]==1 && b->line6id[2]==0xC)
	{
	  switch(le16(b->family))
	    {
	    case 0:
	      podname+="POD ";
	      switch(le16(b->product))
		{
		case 0x0100: podname+="1 "; break;
		case 0x0200: podname+="Flextone2 "; break;
		case 0x0400: podname+="Pro "; break;
		default: podname+="unknown "; break;
		}
	      break;

	    case 2:
	      podname+="BassPOD ";
	      switch(le16(b->product))
		{
		case 0: podname+="(Pro) "; break;
		default: podname+="unknown "; break;
		}
	      break;

	    default: podname+="unknown "; break;
	    }
	  podname+="Version ";
	  podname+=b->version[0];
	  podname+=b->version[1];
	  podname+='.';
	  podname+=b->version[2];
	  podname+=b->version[3];
	}
      setCaption();
    }
  // Program Patch Dump
  if(len>=7+1+71*2+1 && buf[0]==0xF0 && buf[1]==0 && buf[2]==1 && buf[3]==0xC && buf[4]==1 && buf[5]==1 && buf[6]==1 && buf[7+1+71*2]==0xF7)
    {
      patchversion=buf[7];
      uint8_t dump[71];
      const uint8_t *b=buf+7+1;
      for(int i=0; i<71; ++i)
	{
	  if(*b & 0xF0) return; // assert that upper nibble is 0
	  dump[i]=*b << 4; ++b;
	  if(*b & 0xF0) return; // assert that upper nibble is 0
	  dump[i]+=*b; ++b;
	}
      assert(*b == 0xF7);
      // switches
      parsecontroller(25, dump[0]?127:0);
      parsecontroller(26, dump[1]?127:0);
      parsecontroller(27, dump[2]?127:0);
      parsecontroller(28, dump[3]?127:0);
      parsecontroller(50, dump[4]?127:0);
      parsecontroller(36, dump[5]?127:0);
      parsecontroller(22, dump[6]?127:0);
      parsecontroller(73, dump[7]?127:0);
      // preamp
      parsecontroller(12, dump[8]);
      parsecontroller(13, dump[9]*2);
      parsecontroller(20, dump[10]*2);
      parsecontroller(14, dump[11]*2);
      parsecontroller(15, dump[12]*2);
      parsecontroller(16, dump[13]*2);
      parsecontroller(21, dump[14]*2);
      parsecontroller(17, dump[15]*2);
      // noise gate
      parsecontroller(23, dump[16]*127/96); // todo: check this formula, might be inverse
      parsecontroller(24, dump[17]*2);
      // wahwah
      parsecontroller(4, dump[18]);
      parsecontroller(44,dump[19]);
      parsecontroller(45,dump[20]);
      // vol pedal
      parsecontroller(7,dump[22]);
      parsecontroller(46,dump[23]);
      parsecontroller(47,dump[24]?127:0);
      // delay
      {
	//int i=(dump[26]+dump[27]*256+dump[28]*256*256)/6; // todo: this is plain wrong
	int i=(dump[27]*256+dump[28])*32; // todo: this is plain wrong
	printf("%i, %i, %i, %i\n", dump[26], dump[27], dump[28], i);
	parsecontroller(30,(i>>8)&0xFF);
	parsecontroller(62,i&0xFF);
      }
      parsecontroller(32,dump[34]*2);
      parsecontroller(34,dump[36]*2);
      // reverb
      parsecontroller(37,dump[38]?127:0);
      parsecontroller(38,dump[39]*2);
      parsecontroller(39,dump[40]*2);
      parsecontroller(40,dump[41]*2);
      parsecontroller(41,dump[42]*2);
      parsecontroller(18,dump[43]*2);
      // cab sim
      parsecontroller(71,dump[44]);
      parsecontroller(72,dump[45]*2);
      // fx config
      parsecontroller(19,dump[46]);
      parsecontroller(1,dump[47]*2);
      // swell
      parsecontroller(49,dump[48]*2);
      // comp
      parsecontroller(42,dump[48]*22);
      // chorus, flanger
      parsecontroller(51,(dump[48]+dump[49]*256)/50); // todo: check
      parsecontroller(52,(dump[50]+dump[51]*256)/2);
      parsecontroller(53,dump[52]);
      parsecontroller(54,(dump[53]+dump[54]*256)/4);
      // rotary
      parsecontroller(55, dump[48]?127:0);
      parsecontroller(56, (dump[49]+dump[50]*256)/8); // todo: check
      parsecontroller(57, (dump[51]+dump[52]*256)/8);
      // tremolo
      parsecontroller(58, (dump[48]+dump[49]*256)/8); // todo: check
      parsecontroller(59, dump[50]);
      // name
      patchname[0]= static_cast<const char>(dump[55]);
      patchname[1]= static_cast<const char>(dump[56]);
      patchname[2]= static_cast<const char>(dump[57]);
      patchname[3]= static_cast<const char>(dump[58]);
      patchname[4]= static_cast<const char>(dump[59]);
      patchname[5]= static_cast<const char>(dump[60]);
      patchname[6]= static_cast<const char>(dump[61]);
      patchname[7]= static_cast<const char>(dump[62]);
      patchname[8]= static_cast<const char>(dump[63]);
      patchname[9]= static_cast<const char>(dump[64]);
      patchname[10]=static_cast<const char>(dump[65]);
      patchname[11]=static_cast<const char>(dump[66]);
      patchname[12]=static_cast<const char>(dump[67]);
      patchname[13]=static_cast<const char>(dump[68]);
      patchname[14]=static_cast<const char>(dump[69]);
      patchname[15]=static_cast<const char>(dump[70]);
      setCaption();
      // dig out
      parsecontroller(9, dump[35]*2);
    }
  // All Programs Dump
  if(len>=7+1+2556*2+1 && buf[0]==0xF0 && buf[1]==0 && buf[2]==1 && buf[3]==0xC && buf[4]==1 && buf[5]==1 && buf[6]==2 && buf[7+1+2556*2]==0xF7)
    {
      // get patch names
      for(int p=0; p<36; ++p)
	{
	  uint8_t dump[72];
	  const uint8_t *b=buf+7+1+p*71*2;
	  for(int i=0; i<71; ++i)
	    {
	      if(*b & 0xF0) return; // assert that upper nibble is 0
	      dump[i]=*b << 4; ++b;
	      if(*b & 0xF0) return; // assert that upper nibble is 0
	      dump[i]+=*b; ++b;
	    }
	  dump[71]=0;
	  presetname[p+1]=reinterpret_cast<char*>(dump+55);
	  bgbut[p+1]->setText(QString(presetname_prefix[p+1])+presetname[p+1]);
	}
    }

  // search for current preset number
  for(unsigned i=0; i<sizeof_array(presetname); ++i)
    if(presetname[i] == patchname)
      {
	bg->setButton(i);
	break;
      }
}

void QTPOD::setCaption()
{
  mainWidget()->setCaption(QString("QTPOD " QTPOD_VERSION ": ")+podname+" - "+patchname+" Version:"+QString("%1").arg(patchversion));
}

QTPOD::QTPOD(int argc, char **argv) :
  QApplication(argc, argv),
  patchname("0123456789abcdef"),
  patchversion(0),
  podfound(false)
{
  cout << "qtpod " QTPOD_VERSION << endl;

  argc=this->argc();
  argv=this->argv();

  char *midiclient=0;
  int midichannel=0;
  int c;
  while((c=getopt(argc, argv, "c:m:")) >= 0)
    switch(c)
      {
      case 'c':
	midiclient=optarg;
	break;

      case 'm':
	midichannel=strtoul(optarg,0,0);
	if(midichannel<0) midichannel=0;
	break;

      default:
	cout << endl;
	cout << "usage: qtpod [-m midichannel] [-c alsa midi client]" << endl;
	cout << "alsa midi client is given as clientID:portNum (two integers). You can query" << endl;
	cout << "your system with 'aconnect -i -l'." << endl;
	throw "unknown command line option";
      }

  // init midi
  int midifd;
  if((midifd=midi_init("QTPod", midichannel, midiclient)) < 0)
     throw "could not init midi";

  QSocketNotifier *qsn=new QSocketNotifier(midifd, QSocketNotifier::Read, this, "socket notifier");
  connect(qsn, SIGNAL(activated(int)), this, SLOT(midievent(int)));

  // init GUI

#define SLIDER(box, widget, name) { \
 QHBox *bx=new QHBox(box); \
 widget=new QSlider(0,127,16,0,Qt::Horizontal, bx); \
 connect(widget, SIGNAL(valueChanged(int)), this, SLOT(widget ## _slot(int))); \
 new QLabel(name, bx); \
}

#define CHECKBOX(box, widget, name) \
widget=new QCheckBox(name, box); \
connect(widget, SIGNAL(toggled(bool)), this, SLOT(widget ## _slot(bool)));

#define GROUPBOXMARGIN 5

  QScrollView *sv=new QScrollView(0, "sv");
#define MAINWIDGET sv
  sv->setResizePolicy(QScrollView::AutoOneFit);

  QVBox *mainbox=new QVBox(sv->viewport(), "mainbox");
  sv->addChild(mainbox);
  {
    QGroupBox *switchbox=new QGroupBox(8, Qt::Horizontal, "Switches", mainbox, "switch box");
    switchbox->setInsideMargin(GROUPBOXMARGIN);
    CHECKBOX(switchbox, sw_distortion, "Distortion");
    CHECKBOX(switchbox, sw_drive, "Drive");
    CHECKBOX(switchbox, sw_eq, "Equalizer");
    CHECKBOX(switchbox, sw_delay, "Delay");
    CHECKBOX(switchbox, sw_tremolo, "Tremolo,Rotary,Chorus,Flanger");
    CHECKBOX(switchbox, sw_reverb, "Reverb");
    CHECKBOX(switchbox, sw_noise, "NoiseGate");
    CHECKBOX(switchbox, sw_bright, "Bright");
    sw_bright->setEnabled(false);
    // disable vertical expansion
    QSizePolicy sp=switchbox->sizePolicy();
    sp.setVerData(QSizePolicy::Fixed);
    switchbox->setSizePolicy(sp);

    QHBox *querbox=new QHBox(mainbox);
    {
      QVBox *box0=new QVBox(querbox);
      {
	/////
	const unsigned presetsize=sizeof_array(presetname_prefix);
	bg=new QButtonGroup(presetsize/2, Qt::Vertical, "Presets", box0, "preset button group");
	bg->setExclusive(true);
	bg->setInsideMargin(GROUPBOXMARGIN);
	connect(bg, SIGNAL(clicked(int)), this, SLOT(newpreset(int)));
	for(unsigned i=0; i!=presetsize; ++i)
	  bgbut[i]=new QRadioButton(presetname_prefix[i], bg, presetname_prefix[i]);

	/////
	taptempo_but=new QPushButton("Tap Tempo", box0);
	connect(taptempo_but, SIGNAL(clicked()), this, SLOT(taptempo_slot()));

	QPushButton *but;

	/////
	QHBox *butbox=new QHBox(box0);
	but=new QPushButton("Load from disk", butbox);
	but=new QPushButton("Save to Disk", butbox);

	/////
	butbox=new QHBox(box0);
	but=new QPushButton("Load All from Disk", butbox);
	but=new QPushButton("Save All to Disk", butbox);

      }

      box0=new QVBox(querbox);
      {
	/////
	QGroupBox *preampbox=new QGroupBox(1, Qt::Horizontal, "Preamp", box0, "preamp box");
	preampbox->setInsideMargin(GROUPBOXMARGIN);
	preamp_ampmodel=new QComboBox(false, preampbox);
	for(unsigned i=0; i!=sizeof(ampname)/sizeof(ampname[0]); ++i)
	  preamp_ampmodel->insertItem(ampname[i]);
	connect(preamp_ampmodel, SIGNAL(activated(int)), this, SLOT(preamp_model_slot(int)));

	SLIDER(preampbox, preamp_drive, "Drive");
	SLIDER(preampbox, preamp_drive2, "Drive2");
	SLIDER(preampbox, preamp_bass, "Bass");
	SLIDER(preampbox, preamp_mid, "Mid");
	SLIDER(preampbox, preamp_treble, "Treble");
	SLIDER(preampbox, preamp_presence, "Presence");
	SLIDER(preampbox, preamp_chanvol, "ChanVol");
	SLIDER(preampbox, dig_gain, "DigGain");

	/////
	QGroupBox *cabbox=new QGroupBox(1, Qt::Horizontal, "Cabinet", box0, "cab box");
	cabbox->setInsideMargin(GROUPBOXMARGIN);
	cab_type=new QComboBox(false, cabbox);
	for(unsigned i=0; i!=sizeof(cabname)/sizeof(cabname[0]); ++i)
	  cab_type->insertItem(cabname[i]);
	connect(cab_type, SIGNAL(activated(int)), this, SLOT(cab_type_slot(int)));
	SLIDER(cabbox, cab_air, "Air");

	/////
	QGroupBox *fxbox=new QGroupBox(1, Qt::Horizontal, "FX Config", box0, "fx box");
	fxbox->setInsideMargin(GROUPBOXMARGIN);
	fx_type=new QComboBox(false, fxbox);
	for(unsigned i=0; i!=sizeof(fxname)/sizeof(fxname[0]); ++i)
	  fx_type->insertItem(fxname[i]);
	connect(fx_type, SIGNAL(activated(int)), this, SLOT(fx_type_slot(int)));
	SLIDER(fxbox, fx_tweak, "Tweak");

	/////
	QGroupBox *fxxbox=new QGroupBox(1, Qt::Horizontal, "Misc FX", box0, "effects box");
	fxxbox->setInsideMargin(GROUPBOXMARGIN);
	SLIDER(fxxbox, swell_time, "Swell Attack Time");
	SLIDER(fxxbox, comp_ration, "Compression Ratio");

	/////
	QGroupBox *wahbox=new QGroupBox(1, Qt::Horizontal, "WahWah", box0, "wah wah box");
	wahbox->setInsideMargin(GROUPBOXMARGIN);
	CHECKBOX(wahbox, wah_enable, "Enable");
	SLIDER(wahbox, wah_level, "Level");
	SLIDER(wahbox, wah_botfreq, "Bottom Freq");
	SLIDER(wahbox, wah_topfreq, "Top Freq");

	/////
	QGroupBox *ngbox=new QGroupBox(1, Qt::Horizontal, "Noise Gate", box0, "noise gate box");
	ngbox->setInsideMargin(GROUPBOXMARGIN);
	SLIDER(ngbox, ng_threshold, "Theshold");
	SLIDER(ngbox, ng_decay, "Decay Time");
      }

      box0=new QVBox(querbox);
      {
	/////
	QGroupBox *volpedalbox=new QGroupBox(1, Qt::Horizontal, "VolPedal", box0, "vol pedal box");
	volpedalbox->setInsideMargin(GROUPBOXMARGIN);
	SLIDER(volpedalbox, volpedal_level, "Level");
	SLIDER(volpedalbox, volpedal_min, "Min");
	CHECKBOX(volpedalbox, volpedal_pos, "Pos: Pre/Post");

	/////
	QGroupBox *chorusbox=new QGroupBox(1, Qt::Horizontal, "Chorus / Flanger", box0, "chorus box");
	chorusbox->setInsideMargin(GROUPBOXMARGIN);
	SLIDER(chorusbox, chorus_speed, "Speed");
	SLIDER(chorusbox, chorus_depth, "Depth");
	SLIDER(chorusbox, chorus_feedback, "Feedback");
	SLIDER(chorusbox, chorus_delay, "PreDelay");

	/////
	QGroupBox *rotarybox=new QGroupBox(1, Qt::Horizontal, "Rotary Speaker", box0, "rotary box");
	rotarybox->setInsideMargin(GROUPBOXMARGIN);
	CHECKBOX(rotarybox, rotary_curr, "Fast/Slow");
	SLIDER(rotarybox, rotary_fast, "FastSpeed");
	SLIDER(rotarybox, rotary_slow, "SlowSpeed");

	/////
	QGroupBox *tremolobox=new QGroupBox(1, Qt::Horizontal, "Tremolo", box0, "tremolo box");
	tremolobox->setInsideMargin(GROUPBOXMARGIN);
	SLIDER(tremolobox, tremolo_speed, "Speed");
	SLIDER(tremolobox, tremolo_depth, "Depth");

	/////
	QGroupBox *delaybox=new QGroupBox(1, Qt::Horizontal, "Delay", box0, "delay box");
	delaybox->setInsideMargin(GROUPBOXMARGIN);
	{
	  QHBox *bx=new QHBox(delaybox);
	  delay_time1=new QSlider(0,128*128-1,128*16,0,Qt::Horizontal, bx);
	  connect(delay_time1, SIGNAL(valueChanged(int)), this, SLOT(delay_time1_slot(int)));
	  new QLabel("Time", bx);
	}
	SLIDER(delaybox, delay_feedback1, "Feedback");
	SLIDER(delaybox, delay_level1, "Level");

	/////
	QGroupBox *reverbbox=new QGroupBox(1, Qt::Horizontal, "Reverb", box0, "reverb box");
	reverbbox->setInsideMargin(GROUPBOXMARGIN);

	reverb_type=new QButtonGroup(1, Qt::Vertical, "Type", reverbbox);
	reverb_type->setExclusive(true);
	reverb_type->setInsideMargin(GROUPBOXMARGIN);
	connect(reverb_type, SIGNAL(clicked(int)), this, SLOT(reverb_type_slot(int)));
	new QRadioButton("Spring", reverb_type);
	new QRadioButton("Hall", reverb_type);

	SLIDER(reverbbox, reverb_decay, "Decay");
	SLIDER(reverbbox, reverb_tone, "Tone");
	SLIDER(reverbbox, reverb_diffusion, "Diffusion");
	SLIDER(reverbbox, reverb_density, "Density");
	SLIDER(reverbbox, reverb_level, "Level");
      }

    }
  }
#undef SLIDER

  MAINWIDGET->show();
  setMainWidget(MAINWIDGET);
#undef MAINWIDGET

  // set icon for all top level widgets
#include "pod.xpm"
  QWidgetList *list = QApplication::topLevelWidgets();
  QWidgetListIt it( *list );  // iterate over the widgets
  while (it.current()) {   // for each top level widget...
    it.current()->setIcon(pod);
    ++it;
  }
  delete list;

  QTimer::singleShot(TIMER_DELAY, this, SLOT(sendpatchquery()));

#if 1
  QTimer::singleShot(1000, this, SLOT(restore_geometry_slot()));
#else
  restore_geometry_slot();
#endif
}

void QTPOD::restore_geometry_slot()
{
  QWidget *mw=mainWidget();

  QSettings settings;
  int w=settings.readNumEntry("/qtpod/size/width"), h=settings.readNumEntry("/qtpod/size/height");
  if(w>50 && h>50)
    mw->resize(w,h);
  int x=settings.readNumEntry("/qtpod/pos/x"), y=settings.readNumEntry("/qtpod/pos/y");
  if(x>=0 && x<QApplication::desktop()->width()-50 && y>=0 && y<QApplication::desktop()->height()-50)
    mw->move(x,y);
}

QTPOD::~QTPOD()
{
  QSettings settings;

  QSize s = mainWidget()->size();
  settings.writeEntry("/qtpod/size/width", s.width());
  settings.writeEntry("/qtpod/size/height", s.height());

  QPoint p = mainWidget()->pos();
  settings.writeEntry("/qtpod/pos/x", p.x());
  settings.writeEntry("/qtpod/pos/y", p.y());
}

void QTPOD::updatebrightsw(int i)
{
  preamp_drive2->setEnabled(i==4);
  switch(i)
    {
    case 1:
    case 2:
    case 3:
    case 4:
    case 10:
    case 16:
    case 22:
    case 24:
      sw_bright->setEnabled(true); break;
    default:
      sw_bright->setEnabled(false); break;
    }
}

void QTPOD::newpreset(int id)
{
  do_programchange(id);
}

void QTPOD::sendpatchquery()
{
  uint8_t buf [] = { 0xF0, 0x00, 0x01, 0x0C, 0x01, 0x00, 0x01, 0xF7 };
  do_sysex(buf, sizeof(buf));

  if(!podfound)
    {
      QTimer::singleShot(TIMER_DELAY, this, SLOT(sendallquery()));
      QTimer::singleShot(TIMER_DELAY*2, this, SLOT(sendpodquery()));
    }
}
void QTPOD::sendpodquery()
{
  uint8_t buf[] = { 0xF0, 0x7E, 0x7F, 0x06, 0x01, 0xF7 };
  do_sysex(buf, sizeof(buf));
}
void QTPOD::sendallquery()
{
  uint8_t buf [] = { 0xF0, 0x00, 0x01, 0x0C, 0x01, 0x00, 0x02, 0xF7 };
  do_sysex(buf, sizeof(buf));
}

void QTPOD::taptempo(int i)
{
  taptempo_but->setPaletteBackgroundColor(QColor(255,0,0));
  QTimer::singleShot(100, this, SLOT(taptempo_resetbg()));
}
void QTPOD::taptempo_resetbg()
{
  taptempo_but->unsetPalette();
}
void QTPOD::taptempo_slot()
{
  do_controller(64, 127);
}

void QTPOD::sw_distortion_slot(bool b)
{
  do_controller(25, b?127:0);
}
void QTPOD::sw_drive_slot(bool b)
{
  do_controller(26, b?127:0);
}
void QTPOD::sw_eq_slot(bool b)
{
  do_controller(27, b?127:0);
}
void QTPOD::sw_delay_slot(bool b)
{
  do_controller(28, b?127:0);
}
void QTPOD::sw_tremolo_slot(bool b)
{
  do_controller(50, b?127:0);
}
void QTPOD::sw_reverb_slot(bool b)
{
  do_controller(36, b?127:0);
}
void QTPOD::sw_noise_slot(bool b)
{
  do_controller(22, b?127:0);
}
void QTPOD::sw_bright_slot(bool b)
{
  do_controller(73, b?127:0);
}

void QTPOD::dig_gain_slot(int i)
{
  do_controller(9,i);
}

void QTPOD::swell_time_slot(int i)
{
  do_controller(49,i);
}
void QTPOD::comp_ration_slot(int i)
{
  do_controller(42,i);
}
void QTPOD::chorus_speed_slot(int i)
{
  do_controller(51,i);
}
void QTPOD::chorus_depth_slot(int i)
{
  do_controller(52,i);
}
void QTPOD::chorus_feedback_slot(int i)
{
  do_controller(53,i);
}
void QTPOD::chorus_delay_slot(int i)
{
  do_controller(54,i);
}
void QTPOD::rotary_curr_slot(bool i)
{
  do_controller(55,i?127:0);
}
void QTPOD::rotary_fast_slot(int i)
{
  do_controller(56,i);
}
void QTPOD::rotary_slow_slot(int i)
{
  do_controller(57,i);
}
void QTPOD::tremolo_speed_slot(int i)
{
  do_controller(58,i);
}
void QTPOD::tremolo_depth_slot(int i)
{
  do_controller(59,i);
}

void QTPOD::fx_type_slot(int i)
{
  do_controller(19,i);
}
void QTPOD::fx_tweak_slot(int i)
{
  do_controller(1,i);
}

void QTPOD::cab_type_slot(int i)
{
  do_controller(71,i);
}
void QTPOD::cab_air_slot(int i)
{
  do_controller(72,i);
}

void QTPOD::reverb_type_slot(int i)
{
  do_controller(37,i?127:0);
}
void QTPOD::reverb_decay_slot(int i)
{
  do_controller(38,i);
}
void QTPOD::reverb_tone_slot(int i)
{
  do_controller(39,i);
}
void QTPOD::reverb_diffusion_slot(int i)
{
  do_controller(40,i);
}
void QTPOD::reverb_density_slot(int i)
{
  do_controller(41,i);
}
void QTPOD::reverb_level_slot(int i)
{
  do_controller(18,i);
}

void QTPOD::delay_time1_slot(int i)
{
  do_controller(30,(i>>7)&0x7F);
  do_controller(62,i&0x7F);
}
void QTPOD::delay_feedback1_slot(int i)
{
  do_controller(32,i);
}
void QTPOD::delay_level1_slot(int i)
{
  do_controller(34,i);
}

void QTPOD::volpedal_level_slot(int i)
{
  do_controller(7,i);
}
void QTPOD::volpedal_min_slot(int i)
{
  do_controller(46,i);
}
void QTPOD::volpedal_pos_slot(bool i)
{
  do_controller(47,i?127:0);
}

void QTPOD::wah_level_slot(int i)
{
  do_controller(4,i);
#if 0
  // this does not work
  if(!wah_enable->isChecked())
    wah_enable->setChecked(true);
#endif
}
void QTPOD::wah_botfreq_slot(int i)
{
  do_controller(44,i);
}
void QTPOD::wah_topfreq_slot(int i)
{
  do_controller(45,i);
}
void QTPOD::wah_enable_slot(bool b)
{
  do_controller(43,b?127:0);
}

void QTPOD::ng_threshold_slot(int i)
{
  do_controller(23,i);
}
void QTPOD::ng_decay_slot(int i)
{
  do_controller(24,i);
}

void QTPOD::preamp_model_slot(int i)
{
  do_controller(12,i);
  updatebrightsw(i);
}
void QTPOD::preamp_drive_slot(int i)
{
  do_controller(13,i);
}
void QTPOD::preamp_drive2_slot(int i)
{
  do_controller(20,i);
}
void QTPOD::preamp_bass_slot(int i)
{
  do_controller(14,i);
}
void QTPOD::preamp_mid_slot(int i)
{
  do_controller(15,i);
}
void QTPOD::preamp_treble_slot(int i)
{
  do_controller(16,i);
}
void QTPOD::preamp_presence_slot(int i)
{
  do_controller(21,i);
}
void QTPOD::preamp_chanvol_slot(int i)
{
  do_controller(17,i);
}

int main(int argc, char **argv)
{
  int ret=1;
  QTPOD *p=0;
  try {
    p=new QTPOD(argc, argv);
    ret=p->exec();
  } catch(std::exception& e) {
    cerr << endl << "Exception caught: " << e.what() << endl;
  } catch(char const * s) {
    cerr << endl << "Exception caught: " << s << endl;
  } catch(...) {
    cerr << endl << "unknown exception caught" << endl;
  }
  if(p)
    {
      p->quit();
      delete p;
    }
  return ret;
}
